{-# LANGUAGE NamedFieldPuns #-}
module Main (main) where

import qualified Data.ByteString.Lazy as LazyBS
import Data.Foldable
import Data.Function ((&))
import qualified Data.Map.Strict as Map
import Data.Set (Set)
import qualified Data.Set as Set
import Data.Text (Text)
import Misc
import qualified System.FilePath.Find as Find
import YAML ((##=), (.!=), (.:))
import qualified YAML
import System.Environment
import qualified Data.Text.IO as Text
import qualified Data.Text as Text


-- |
-- @all-tags dir yaml make@:
--
-- * reads all posts in @dir@
-- * collects all tags, and assigns them slugs ('getTags')
-- * writes a list of {tag, slug, count} to @yaml@ ('makeYAML')
-- * writes recipes to build the tag pages to @make@ ('makeMake')
--
-- the program fails on slug collision, because the alternative is for the page
-- generation to consult the global list for slugs, and i haven't done that
-- currently
main :: IO ()
main = do
  Opts {dir, yaml, make} <- getOptions
  files <- Find.findL True (pure True) (Find.extension Find.==? ".md") dir
  tags  <- traverse getTags files
  LazyBS.writeFile yaml $ makeYAML tags
  Text.writeFile make $ makeMake tags

-- | reads tags from a single file
getTags :: FilePath -> IO (Set Text)
getTags file = do
  yaml <- YAML.readHeader file
  list <- unwrap file $ YAML.parseEither $
    yaml & YAML.withMap "yaml header" \m -> m .: "tags" .!= []
  pure $ Set.fromList list

-- | generates a yaml document from a set of tags from each file.
-- example:
--
-- @
-- ---
-- title: all tags
-- all-tags:
--   - name: tag1
--     slug: tag1
--     count: 2
--   - name: tag2
--     slug: tag2
--     count: 7
-- ...
-- @
makeYAML :: [Set Text] -> LazyBS.ByteString
makeYAML tags = "---\n" <> yaml <> "\n...\n" where
  yaml = YAML.encode1 $ YAML.obj
    [("title"    ##= YAML.str "all tags"),
     ("all-tags" ##= collate tags)]

-- | generates a makefile to include that generates the index and individual
-- tag pages. example:
--
-- @
-- build: $(BUILDDIR)/tag1.html $(BUILDDIR)/tag2.html
-- $(TMPDIR)/index.md : $(POSTS) $(POST_LISTS)
-- 	@echo "[post-lists] index"
-- 	$(POST_LISTS) --out $@ $(POSTSDIR) "all posts"
-- $(TMPDIR)/tag-ats.md : $(POSTS) $(POST_LISTS)
-- 	@echo "[post-lists] tag-tag1"
-- 	$(POST_LISTS) --tag "tag1" --out $@ $(POSTSDIR) "posts tagged ‘tag1’"
-- $(TMPDIR)/tag-ats.md : $(POSTS) $(POST_LISTS)
-- 	@echo "[post-lists] tag-tag2"
-- 	$(POST_LISTS) --tag "tag2" --out $@ $(POSTSDIR) "posts tagged ‘tag2’"
-- @
--
-- the generated rules depend on these make variables:
--
-- * @BUILDDIR@ (output of the final site)
-- * @TMPDIR@ (intermediate file hole)
-- * @POSTSDIR@ (directory containing the post @.md@ files)
-- * @POSTS@ (list of posts (**including** @POSTSDIR@, because that is what
--   @find@ returns)
-- * @POST_LISTS@ (path to the @post-lists@ executable)
makeMake :: [Set Text] -> Text
makeMake tags' = Text.unlines $ build : allPosts : map makeRule tags where
  build = Text.unwords $
    "build:" : ["$(BUILDDIR)/" <> t <> ".html" | t <- map slug' tags]
  makeRule' opt title file =
    "$(TMPDIR)/" <> file <> ".md : $(POSTS) $(POST_LISTS)\n\
    \\t@echo \"[post-lists] " <> file <> "\"\n\
    \\t$(POST_LISTS) " <> opt <> " --out $@ \\\n\
    \\t  $(POSTSDIR) \"" <> title <> "\""
  allPosts = makeRule' "" "all posts" "index"
  makeRule t =
    makeRule' ("--tag \"" <> name t <> "\"")
              ("posts tagged ‘" <> name t <> "’")
              (slug' t)
  slug' (Tag {slug}) = "tag-" <> slug
  tags = collate tags'

-- | info about a tag
data Tag =
  Tag {
    name :: !Text, -- ^ display name
    slug :: !Text, -- ^ name in urls
    count :: !Int  -- ^ how many posts with this tag
  } deriving Show

instance YAML.ToYAML Tag where
  toYAML (Tag {name, slug, count}) = YAML.obj $
    [("name" ##= name), ("slug" ##= slug), ("count" ##= count)]

-- | counts occurrences of each tag, and checks they all have distinct slugs
collate :: [Set Text] -> [Tag]
collate =
  toList . foldl' add1 mempty . foldMap toList
 where
  add1 tags name = Map.alter (add1' tag) (slug tag) tags
    where tag = makeTag name

  add1' new Nothing = Just new
  add1' new (Just old)
    | name new == name old = Just (old {count = succ (count old)})
    | otherwise = error $ "slug collision between " <> show (name old) <>
                          " and " <> show (name new)

  makeTag name = Tag {name, slug = makeSlug name, count = 1}

-- | command line options
data Options =
  Opts {
    dir :: !FilePath,  -- ^ first argument: dir containing posts
    yaml :: !FilePath, -- ^ second argument: yaml output file
    make :: !FilePath  -- ^ third argument: make output file
  }

getOptions :: IO Options
getOptions = do
    args <- getArgs
    prog <- getProgName
    case args of
      [dir, yaml, make] -> pure $ Opts {dir, yaml, make}
      _ -> fail $
        "usage: " <> prog <> " DIR YAML MAKE ---\n\
        \  get all tags from posts in DIR and put the results in the given files"
